part of wrike;

typedef dynamic MapToObjectConvertor(Map values);

class WrikeApi {

  final WrikeConfig _config;
  final WrikeOAuth oauth2;


  WrikeApi(WrikeConfig config)
      : this._config = config,
        this.oauth2 = new WrikeOAuth(config) {
  }

  Future<WrikeToken> authorize() {
    return oauth2.authorize().then((token) {
      _config.tokenStorage.setToken(token);
      return token;
    });
  }

  /*
   * --------------------------- Attachment ---------------------------
   */
  /**
   * Add an Attachment to a task
   */
  Future<Attachment> addAttach(String taskId, String filename, dataBase64) {
    Map<String, String> headers = {
      'X-File-Name': filename
    };
    return oauth2.post('tasks/$taskId/attachments', null, headers, dataBase64).then(_parseAttachment);
  }

  /**
   *  Update previously uploaded Attachment with new version.
   */
  Future<Attachment> updateAttach(String attachmentId, String filename, dataBase64) {
    Map<String, String> headers = {
      'X-File-Name': filename
    };
    return oauth2.put('attachments/${attachmentId}', null, headers, dataBase64).then(_parseAttachment);
  }

  /**
   * Delete Attachment by ID.
   */
  Future<Map> deleteAttach(String attachmentId) {
    return oauth2.delete('attachments/${attachmentId}', null);
  }

  /**
   * Returns URL that can be used to download Attachment from Wrike or external service.
   */
  Future<AttachmentUrl> getAttachUrl(String attachmentId) {
    return oauth2.get('/attachments/${attachmentId}/url', null).then(_parseAttachmentUrl);
  }


  /*
   * --------------------------- TASK ---------------------------
   */


  Map createGetTasksParams(List<TaskFields> fields, TaskStatus status, {String textSearch, String permalink, int limit}) {
    var params = {};
    if (fields != null) {
      params["fields"] = JSON.encode(fields);
    }
    if (status != null) {
      params["status"] = status.toString();
    }
    if (textSearch != null) {
      params["textSearch"] = JSON.encode([textSearch]);
    }
    if (permalink != null) {
      params["permalink"] = permalink;
    }
    if (limit != null) {
      params["limit"] = limit.toString();
    }

    return params;
  }

  /**
     * Search over tasks in the folder.
     */
  Future<List<Task>> getTasksByFolder(String folderId, {List<TaskFields> fields, TaskStatus status, String textSearch, String permalink, int limit}) {
    return _getList('/folders/$folderId/tasks', Task.fromJSON, createGetTasksParams(fields, status, textSearch: textSearch, permalink: permalink, limit: limit));
  }

  /**
     * Search over all tasks in all accounts.
     */
  Future<List<Task>> getTasks({List<TaskFields> fields, TaskStatus status, String textSearch, String permalink, int limit}) {
    return _getList('/tasks', Task.fromJSON, createGetTasksParams(fields, status, textSearch: textSearch, permalink: permalink, limit: limit));
  }

  /**
     * Returns complete information about the task.
     */
  Future<Task> getTask(String taskId) {
    return _get('/tasks/$taskId', Task.fromJSON, {});
  }

  /**
      * Returns complete information about multiple tasks.
    */
  Future<List<Task>> getTasksByIds(List<String> taskIds) {
    return _getList('/tasks/${taskIds.join(',')}', Task.fromJSON, {});
  }

  /**
   * Create task in folder. You can specify rootFolderId to create task in user's account root.
   */
  Future<Task> createTask(String folderId, Task task) {
    Map params = _createTaskParams(task);
    return oauth2.post('/folders/$folderId/tasks', params).then(_parseTask);
  }

  Map _createTaskParams(Task task) {
    Map<String, String> params = {};
    if (task.title != null) {
      params["title"] = task.title;
    }
    if (task.dates != null) {
      params["dates"] = JSON.encode(task.dates);
    }
    if (task.description != null) {
      params["description"] = task.description;
    }
    if (task.status != null) {
      params["status"] = task.status.toString();
    }
    if (task.parentIds != null) {
      params["parents"] = JSON.encode(task.parentIds);
    }
    if (task.importance != null) {
      params["importance"] = task.importance.toString();
    }
    if (task.sharedIds != null) {
      params["shareds"] = JSON.encode(task.sharedIds);
    }
    if (task.responsibleIds != null) {
      params["responsibles"] = JSON.encode(task.responsibleIds);
    }
    return params;
  }

  Map _updateTaskParams(String title, String description, status, importance, List<String> addParents, 
                        List<String> removeParents, List<String> addShareds, List<String> removeShareds, List<String> addResponsibles, List<String> removeResponsibles,
                        TaskDates dates) {
    Map<String, String> params = {};
    if (title != null) {
      params["title"] = title;
    }
    if (description != null) {
      params["description"] = description;
    }
    if (dates != null){
      params["dates"] = JSON.encode(dates);
    }
    if (status != null) {
      params["status"] = status.toString();
    }
    if (importance != null) {
      params["importance"] = importance.toString();
    }
    if (addParents != null && addParents.length != 0) {
      params["addParents"] = JSON.encode(addParents);
    }
    if (removeParents != null && removeParents.length != 0) {
      params["removeParents"] = JSON.encode(removeParents);
    }
    if (addShareds != null && addShareds.length != 0) {
      params["addShareds"] = JSON.encode(addShareds);
    }
    if (removeShareds != null && removeShareds.length != 0) {
      params["removeShareds"] = JSON.encode(removeShareds);
    }
    if (addResponsibles != null && addResponsibles.length != 0) {
      params["addResponsibles"] = JSON.encode(addResponsibles);
    }
    if (removeResponsibles != null && removeShareds.length != 0) {
      params["removeResponsibles"] = JSON.encode(removeResponsibles);
    }
    return params;
  }

  /**
   * Update task.
   */
  Future<Task> updateTask(String taskId, {String title, String description, TaskStatus status, TaskImportance importance, List<String> addParents, List<String> removeParents, List<String> addShareds, List<String> removeShareds
    , List<String> addResponsibles, List<String> removeResponsibles, TaskDates dates}) {
    Map params = _updateTaskParams(title, description, status, importance, addParents, removeParents, addShareds, removeShareds,addResponsibles,removeResponsibles,dates);
    return oauth2.put('/tasks/${taskId}', params).then(_parseTask);
  }

  /**
   * Delete task by Id.
   */
  Future<Task> deleteTask(String taskId) {
    return oauth2.delete('/tasks/$taskId', null).then(_parseTask);
  }

  /*
   * --------------------------- FOLDER ---------------------------
   */

  /**
   * Create a folder within a folder. Specify virtual rootFolderId in order to create a folder in the account root.
   */
  Future<Folder> createFolder(String rootFolderId, Folder folder) {
    var params = _createFolderParams(folder);
    return oauth2.post('/folders/$rootFolderId/folders', params).then(_parseFolder);
  }

  /**
   * Update folder
   */
  Future<Folder> updateFolder(String folderId, {String title, String description, List<String> addParents, List<String> removeParents, List<String> addShareds, List<String> removeShareds}) {
    var params = _updateFolderParams(title, description, addParents, removeParents, addShareds, removeShareds);
    return oauth2.put('folders/${folderId}', params).then(_parseFolder);
  }

  /**
   * Returns complete information about the folder.
   */
  Future<Folder> getFolder(String folderId) {
    return oauth2.get('folders/${folderId}', {}).then(_parseFolder);
  }

  /**
   * Deletes folder and all descendant folders and tasks unless they have parents outside of deletion scope
  */
  Future<Folder> deleteFolder(String folderId) {
    return oauth2.delete('/folders/$folderId', null).then(_parseFolder);
  }

  _createFolderParams(Folder folder) {
    Map<String, String> params = {};
    if (folder.title != null) {
      params["title"] = folder.title;
    }
    if (folder.description != null) {
      params["description"] = folder.description;
    }
    if (folder.sharedIds != null) {
      params["shareds"] = JSON.encode(folder.sharedIds);
    }
    return params;
  }

  /**
   Returns a list of tree entries for the account
  */
  Future<List<Folder>> getFoldersByAccountId(String accountId) {
    return _getList('/accounts/$accountId/folders', Folder.fromJSON);
  }
  
  /**
   * Returns a list of tree entries for subtree of this folder. 
   * For root and recycle bin folders, returns subtrees of root and recycle bin tasks respectively.
   */
  Future<List<Folder>> getFoldersByFolder(String folderId) {
      return _getList('/folders/$folderId/folders', Folder.fromJSON);
  }

  Map _updateFolderParams(String title, String description, List<String> addParents, List<String> removeParents, List<String> addShareds, List<String> removeShareds) {
    Map<String, String> params = {};
    params["title"] = title;
    if (description != null) {
      params["description"] = description;
    }
    if (addParents != null && addParents.length != 0) {
      params["addParents"] = JSON.encode(addParents);
    }
    if (removeParents != null && removeParents.length != 0) {
      params["removeParents"] = JSON.encode(removeParents);
    }
    if (addShareds != null && addShareds.length != 0) {
      params["addShareds"] = JSON.encode(addShareds);
    }
    if (removeShareds != null && removeShareds.length != 0) {
      params["removeShareds"] = JSON.encode(removeShareds);
    }
    return params;
  }

  /*
 * --------------------------- CONTACT ---------------------------
 */

  /**
   * List of users and user groups in account
   */
  Future<List<User>> getContactsByAccount(String accountId) {
    return _getList('/accounts/$accountId/contacts', User.fromJSON);
  }

  /**
   * List information about all users in all accessible accounts.
   */
  Future<List<User>> getContacts([bool me = false]) {
    return _getList('/contacts', User.fromJSON,me?{'me':''}:null);
  }

  /*
 * --------------------------- Account ---------------------------
 */

  /**
   * Returns all accounts to which user has access.
   */
  Future<List<Account>> getAccounts() {
    return _getList('/accounts', Account.fromJSON);
  }

  /**
   * Returns specified account.
   */
  Future<Account> getAccount(String accountId) {
    return _get('/accounts/$accountId', Account.fromJSON);
  }

  /**
    *  Returns list of entries required to build a folder tree for all accounts.
    */
  Future<List<Folder>> getFolders() {
    return _getList('/folders', Folder.fromJSON);
  }

  /*
   * --------------------------- COMMENT ---------------------------
   */

  /**
   * Create comment in task.
   */
  Future<Comment> createComment(String taskId, String text) {
    var params = _createCommentParams(text);
    return oauth2.post('/tasks/$taskId/comments', params).then(_parseComment);
  }


  /**
   * Update Comment by ID. A comment is available for updates only during the 5 minutes after creation.
   */
  Future<Comment> updateComment(String commentId, String text) {
    var params = _createCommentParams(text);
    return oauth2.put('/comments/${commentId}', params).then(_parseComment);
  }

  /**
   * Get task comments.
   */
  Future<List<Comment>> getComments(String taskId) {
    return _getList('/tasks/${taskId}/comments', _parseComment);
  }

  Map<String, String> _createCommentParams(String text) {
    Map<String, String> params = {};
    params["text"] = text;
    return params;
  }

  /*
   * --------------------------- PARSER ---------------------------
   */

  Future<dynamic> _getList(String url, MapToObjectConvertor mapper, [Map<String, String> params = null]) {
    return oauth2.get(url, params).then((v) => v["data"].map(mapper).toList());
  }
  Future<dynamic> _get(String url, MapToObjectConvertor mapper, [Map<String, String> params = null]) {
    return oauth2.get(url, params).then((v) => mapper(v["data"][0]));
  }


  Task _parseTask(result) => Task.fromJSON(result["data"][0]);
  Folder _parseFolder(result) => Folder.fromJSON(result["data"][0]);
  Attachment _parseAttachment(result) => Attachment.fromJSON(result["data"][0]);
  AttachmentUrl _parseAttachmentUrl(result) => AttachmentUrl.fromJSON(result["data"][0]);
  Comment _parseComment(Map v) => Comment.fromJSON(v["data"][0]);
}
