part of wrike;

class WrikeOAuth {
  static final Logger _log = new Logger('WrikeOAuth');
  final WrikeConfig _config;
  StreamController<Exception> _errorStream = new StreamController();
  Stream<Exception> get onError => _errorStream.stream;

  WrikeOAuth(WrikeConfig this._config);

  Future<Map> delete(String url, Map<String, String> params, [Map headers]) {
    url = _getUrl(url);
    return _checkAndRefreshToken().then((WrikeToken tokenData) {
      return _config.httpAdapter.delete(url, params, _injectTokenToHeaders(tokenData, headers));
    }).catchError(_errorToStream);
  }

  Future<Map> get(String url, Map<String, String> params, [Map headers]) {
    url = _getUrl(url);
    return _checkAndRefreshToken().then((WrikeToken tokenData) {
      return _config.httpAdapter.get(url, params, _injectTokenToHeaders(tokenData, headers));
    }).catchError(_errorToStream);
  }

  Future<Map> post(String url, Map<String, String> params, [Map<String, String> headers, body]) {
    url = _getUrl(url);
    return _checkAndRefreshToken().then((WrikeToken tokenData) {
      return _config.httpAdapter.post(url, params, _injectTokenToHeaders(tokenData, headers), body: body);
    }).catchError(_errorToStream);
  }

  Future<Map> put(String url, Map<String, String> params, [Map<String, String> headers, body]) {
    url = _getUrl(url);
    return _checkAndRefreshToken().then((WrikeToken tokenData) {
      return _config.httpAdapter.put(url, params, _injectTokenToHeaders(tokenData, headers), body: body);
    }).catchError(_errorToStream);
  }

  Future<WrikeToken> authorize() {
    if (_config.authorizeMethod != null) {
      return _config.authorizeMethod(_config, this);
    }
    return _config.httpAdapter.post(_config.tokenEndpoint.toString(), {
      "client_id": _config.clientId,
      "client_secret": _config.clientSecret,
      "grant_type": "password",
      "username": _config.user,
      "password": _config.password
    }, null).then(parseTokens);
  }

  Future<WrikeToken> parseTokens(Map<String, dynamic> tokenData) {
    var token = new WrikeToken(tokenData["access_token"], tokenData["refresh_token"], tokenData["expires_in"]);
    return _config.tokenStorage.setToken(token).then((_) => token);
  }

  Map<String, String> getTokensParams(code) {
    var params = {
      "client_id": _config.clientId,
      "client_secret": _config.clientSecret,
      "grant_type": 'authorization_code',
      "code": code,
      "redirect_uri": _config.redirectUrl
    };
    if (_config.clientVersion != null && _config.clientVersion.isNotEmpty) {
      params["client_version"] = _config.clientVersion;
    }
    return params;
  }
  String parseAuthorizationCode(String url) {
    var data = new RegExp(r"code=(\w+)").firstMatch(url);
    if (data == null) {
      return null;
    } else {
      return data.group(1);
    }
  }


  _getUrl(urlPath) {
    return _config.API_BASE + urlPath;
  }

  _injectTokenToHeaders(WrikeToken tokenData, Map<String, String> headers) {
    if (headers == null) {
      headers = {};
    }
    headers["Authorization"] = "bearer ${tokenData.access_token}";
    return headers;
  }

  Future<WrikeToken> _refreshFuture = null;

  Future<WrikeToken> _checkAndRefreshToken() {
    return _config.tokenStorage.getToken().then((WrikeToken tokenData) {
      if (tokenData == null) {
        throw new AuthException();
      }
      if (tokenData.isExpired()) {
        if (_refreshFuture != null) {
          return _refreshFuture;
        }
        _refreshFuture = _refreshAccessToken(tokenData);
        return _refreshFuture.then((token) {
          _refreshFuture = null;
          return token;
        }).catchError((error) {
          _refreshFuture = null;
          throw error;
        });
      } else {
        return tokenData;
      }
    });
  }


  _errorToStream(error) {
    if (error is Exception) {
      _errorStream.add(error);
    }
    throw error;
  }



  Future<WrikeToken> _refreshAccessToken(WrikeToken tokenData) {
    if (tokenData == null) {
      throw new AuthException();
    }
    return _config.httpAdapter.post(_getTokensURI(), _getRefreshAccessTokenParams(tokenData.refresh_token), null)
        .then((tokenData) {
          var token = new WrikeToken(tokenData["access_token"], tokenData["refresh_token"], tokenData["expires_in"]);
          return _config.tokenStorage.setToken(token).then((_) {
            return token;
          });
    }, onError: (err) {
      print("Error: $err");
      if (err is RequestException) {
        if (err.statusCode == 0) {
          throw err;
        }
      }
      throw new AuthException();
    });
  }

  String _getTokensURI() {
    return _config.tokenEndpoint.toString();
  }

  _getRefreshAccessTokenParams(refreshToken) {
    var refreshTokenParams = {
      "client_id": _config.clientId,
      "client_secret": _config.clientSecret,
      "grant_type": 'refresh_token',
      "refresh_token": refreshToken
    };
    if (_config.clientVersion != null && _config.clientVersion.isNotEmpty) {
      refreshTokenParams["client_version"] = _config.clientVersion;
    }
    return refreshTokenParams;
  }
}
